@@ -7,36 +7,15 @@ import inspect
 from collections import OrderedDict
 from argparse import Namespace
 from . import _repr_html, _repr_text, _deprecated
+from . import typing as _mtp
 import numpy as np
-from typing import (
-    Dict,
-    Iterable,
-    Any,
-    Union,
-    List,
-    Tuple,
-    Generator,
-    Sized,
-    Optional,
-    Callable,
-    TypeVar,
-    Generic,
-)
+import typing as _tp
 import abc
 from time import monotonic
+import warnings
+import sys
 
-T = TypeVar("T")
-
-
-class Indexable(Iterable, Sized, Generic[T]):
-    """Indexable type for mypy."""
-
-    def __getitem__(self, idx: int) -> T:
-        """Get item at index idx."""
-        ...  # pragma: no cover
-
-
-UserBound = Optional[Indexable[Optional[float]]]
+_ArrayLike = _tp.Sequence
 
 
 class IMinuitWarning(RuntimeWarning):
@@ -57,18 +36,18 @@ class BasicView(abc.ABC):
 
     Derived classes need to implement methods _set and _get to access
     specific properties of the parameter state.
+
+    :meta private:
     """
 
     __slots__ = ("_minuit", "_ndim")
 
-    Key = Union[int, str, slice, List[Union[int, str]]]
-
-    def __init__(self, minuit: Any, ndim: int = 0):
+    def __init__(self, minuit: _tp.Any, ndim: int = 0):
         """Not to be initialized by users."""
         self._minuit = minuit
         self._ndim = ndim
 
-    def __iter__(self) -> Generator:
+    def __iter__(self) -> _tp.Generator:
         """Get iterator over values."""
         for i in range(len(self)):
             yield self._get(i)
@@ -78,14 +57,14 @@ class BasicView(abc.ABC):
         return self._minuit.npar  # type: ignore
 
     @abc.abstractmethod
-    def _get(self, idx: int) -> Any:
+    def _get(self, idx: int) -> _tp.Any:
         return NotImplemented  # pragma: no cover
 
     @abc.abstractmethod
-    def _set(self, idx: int, value: Any) -> None:
-        pass  # pragma: no cover
+    def _set(self, idx: int, value: _tp.Any) -> None:
+        NotImplemented  # pragma: no cover
 
-    def __getitem__(self, key: Key) -> Any:
+    def __getitem__(self, key: _mtp.Key) -> _tp.Any:
         """
         Get values of the view.
 
@@ -95,34 +74,31 @@ class BasicView(abc.ABC):
             If the key is an int or str, return corresponding value.
             If it is a slice, list of int or str, return the corresponding subset.
         """
-        key = _key2index(self._minuit._var2pos, key)
-        if isinstance(key, list):
-            return [self._get(i) for i in key]
-        return self._get(key)
+        index = _key2index(self._minuit._var2pos, key)
+        if isinstance(index, list):
+            return [self._get(i) for i in index]
+        return self._get(index)
 
-    def __setitem__(self, key: Key, value: Any) -> None:
-        """Assign a new value at key, which can be an index, a parameter name, or a slice."""
+    def __setitem__(self, key: _mtp.Key, value: _tp.Any) -> None:
+        """Assign a new value at key, which can be an index, parameter name, or slice."""
         self._minuit._copy_state_if_needed()
-        key = _key2index(self._minuit._var2pos, key)
-        if isinstance(key, list):
+        index = _key2index(self._minuit._var2pos, key)
+        if isinstance(index, list):
             if _ndim(value) == self._ndim:  # support basic broadcasting
-                for i in key:
+                for i in index:
                     self._set(i, value)
             else:
-                if len(value) != len(key):
+                if len(value) != len(index):
                     raise ValueError("length of argument does not match slice")
-                for i, v in zip(key, value):
+                for i, v in zip(index, value):
                     self._set(i, v)
         else:
-            self._set(key, value)
+            self._set(index, value)
 
     def __eq__(self, other: object) -> bool:
         """Return true if all values are equal."""
-        from typing import Iterable, Sized
-
-        if isinstance(other, Iterable) and isinstance(other, Sized):
-            return len(self) == len(other) and all(x == y for x, y in zip(self, other))
-        return NotImplemented
+        a, b = np.broadcast_arrays(self, other)  # type:ignore
+        return _tp.cast(bool, np.all(a == b))
 
     def __repr__(self) -> str:
         """Get detailed text representation."""
@@ -132,16 +108,14 @@ class BasicView(abc.ABC):
         s += ">"
         return s
 
-    def to_dict(self) -> Dict[str, float]:
+    def to_dict(self) -> _tp.Dict[str, float]:
         """Obtain dict representation."""
         return {k: self._get(i) for i, k in enumerate(self._minuit._pos2var)}
 
 
-def _ndim(obj: Iterable) -> int:
-    from typing import Iterable
-
+def _ndim(obj: _tp.Iterable) -> int:
     nd = 0
-    while isinstance(obj, Iterable):
+    while isinstance(obj, _tp.Iterable):
         nd += 1
         for x in obj:
             if x is not None:
@@ -169,6 +143,12 @@ class ErrorView(BasicView):
         return self._minuit._last_state[i].error  # type:ignore
 
     def _set(self, i: int, value: float) -> None:
+        if value <= 0:
+            warnings.warn(
+                "Assigned errors must be positive. "
+                "Non-positive values are replaced by a heuristic."
+            )
+            value = _guess_initial_step(value)
         self._minuit._last_state.set_error(i, value)
 
 
@@ -184,22 +164,26 @@ class FixedView(BasicView):
         else:
             self._minuit._last_state.release(i)
 
+    def __invert__(self) -> list:
+        """Return list with inverted elements."""
+        return [not self._get(i) for i in range(len(self))]
+
 
 class LimitView(BasicView):
     """Array-like view of parameter limits."""
 
-    def __init__(self, minuit: Any):
+    def __init__(self, minuit: _tp.Any):
         """Not to be initialized by users."""
         super(LimitView, self).__init__(minuit, 1)
 
-    def _get(self, i: int) -> Tuple[float, float]:
+    def _get(self, i: int) -> _tp.Tuple[float, float]:
         p = self._minuit._last_state[i]
         return (
             p.lower_limit if p.has_lower_limit else -np.inf,
             p.upper_limit if p.has_upper_limit else np.inf,
         )
 
-    def _set(self, i: int, arg: UserBound) -> None:
+    def _set(self, i: int, arg: _mtp.UserBound) -> None:
         state = self._minuit._last_state
         val = state[i].value
         err = state[i].error
@@ -224,7 +208,7 @@ class LimitView(BasicView):
         state.set_error(i, err)
 
 
-def _normalize_limit(lim: UserBound) -> Tuple[float, float]:
+def _normalize_limit(lim: _mtp.UserBound) -> _tp.Tuple[float, float]:
     if lim is None:
         return (-np.inf, np.inf)
     a, b = lim
@@ -242,12 +226,13 @@ class Matrix(np.ndarray):
     Enhanced Numpy ndarray.
 
     Works like a normal ndarray in computations, but also supports pretty printing in
-    ipython and Jupyter notebooks. Elements can be accessed via indices or parameter names.
+    ipython and Jupyter notebooks. Elements can be accessed via indices or parameter
+    names.
     """
 
     __slots__ = ("_var2pos",)
 
-    def __new__(cls, parameters: Union[Dict, Tuple]) -> Any:
+    def __new__(cls, parameters: _tp.Union[_tp.Dict, _tp.Tuple]) -> _tp.Any:
         """Not to be initialized by users."""
         if isinstance(parameters, dict):
             var2pos = parameters
@@ -256,53 +241,70 @@ class Matrix(np.ndarray):
         else:
             raise TypeError("parameters must be tuple or dict")
         n = len(parameters)
-        obj = super(Matrix, cls).__new__(cls, (n, n))
+        obj = super().__new__(cls, (n, n))
         obj._var2pos = var2pos
         return obj
 
-    def __array_finalize__(self, obj: Any) -> None:
+    def __array_finalize__(self, obj: _tp.Any) -> None:
         """For internal use."""
         if obj is None:
-            self._var2pos = None
+            self._var2pos: _tp.Dict[str, int] = {}
         else:
-            self._var2pos = getattr(obj, "_var2pos", None)
+            self._var2pos = getattr(obj, "_var2pos", {})
 
-    def __getitem__(
+    def __getitem__(  # type:ignore # mypy complains about incompatible signatures
         self,
-        key: Union[Tuple[Union[str, int]], str, int, slice, Iterable[Union[str, int]]],
-    ) -> Any:
+        key: _tp.Union[
+            int,
+            str,
+            _tp.Tuple[_tp.Union[int, str], _tp.Union[int, str]],
+            _tp.Iterable[_tp.Union[int, str]],
+            np.ndarray,
+            slice,
+        ],
+    ) -> np.ndarray:
         """Get matrix element at key."""
-        from typing import Iterable
+        import typing as _tp
 
-        var2pos = self._var2pos or {}
-        if isinstance(key, tuple):  # tuple is special case for __getitem__
-            key = tuple(var2pos.get(k, k) for k in key)
-        elif isinstance(key, str):
-            key = var2pos[key]
-        elif isinstance(key, Iterable) and not isinstance(key, np.ndarray):
-            key = list(var2pos.get(k, k) for k in key)
-            t = super(Matrix, self).__getitem__(key).T
-            return super(Matrix, t).__getitem__(key).T
-        elif isinstance(key, slice):
-            key = slice(var2pos.get(key.start), var2pos.get(key.stop), key.step)
-            return super(Matrix, self).__getitem__((key, key))
-        return super(Matrix, self).__getitem__(key)
+        var2pos = self._var2pos
 
-    def to_dict(self) -> Dict[Tuple[str, str], float]:
+        def trafo(key):
+            if isinstance(key, str):
+                return var2pos[key]
+            if isinstance(key, slice):
+                return slice(trafo(key.start), trafo(key.stop), key.step)
+            if isinstance(key, tuple):
+                return tuple(trafo(k) for k in key)
+            return key
+
+        if isinstance(key, slice):
+            # slice returns square matrix
+            sl = trafo(key)
+            return super().__getitem__((sl, sl))
+        if isinstance(key, (str, tuple)):
+            return super().__getitem__(trafo(key))
+        if isinstance(key, _tp.Iterable) and not isinstance(key, np.ndarray):
+            # iterable returns square matrix
+            index2 = [trafo(k) for k in key]  # type:ignore
+            t = super().__getitem__(index2).T
+            return np.ndarray.__getitem__(t, index2).T
+        return super().__getitem__(key)
+
+    def to_dict(self) -> _tp.Dict[_tp.Tuple[str, str], float]:
         """
         Convert matrix to dict.
 
         Since the matrix is symmetric, the dict only contains the upper triangular matrix.
         """
-        names = tuple(self._var2pos)  # type:ignore
+        names = tuple(self._var2pos)
         d = {}
         for i, pi in enumerate(names):
             for j in range(i, len(names)):
                 pj = names[j]
-                d[pi, pj] = self[i, j]
+                d[pi, pj] = float(self[i, j])
         return d
 
-    def to_table(self) -> Tuple[List[List[str]], Tuple[str, ...]]:
+    def to_table(self) -> _tp.Tuple[_tp.List[_tp.List[str]], _tp.Tuple[str, ...]]:
         """
         Convert matrix to tabular format.
 
@@ -346,6 +348,8 @@ class Matrix(np.ndarray):
 
     def __str__(self):
         """Get user-friendly text representation."""
+        if self.ndim != 2:
+            return repr(self)
         return _repr_text.matrix(self)
 
     def _repr_html_(self):
@@ -404,7 +408,7 @@ class FMin:
 
     def __init__(
         self,
-        fmin: Any,
+        fmin: _tp.Any,
         algorithm: str,
         nfcn: int,
         ngrad: int,
@@ -468,11 +472,11 @@ class FMin:
         """
         Get chi2/ndof of the fit.
 
-        This returns NaN if the cost function is unbinned or does not support
-        reporting the degrees of freedom.
+        This returns NaN if the cost function is unbinned, errordef is not 1,
+        or if the cost function does not report the degrees of freedom.
         """
-        if np.isfinite(self._ndof) and self._ndof > 0:
-            return self.fval / self.errordef / self._ndof
+        if np.isfinite(self._ndof) and self._ndof > 0 and self.errordef == 1:
+            return self.fval / self._ndof
         return np.nan
 
     @property
@@ -651,7 +655,7 @@ class FMin:
     def _repr_html_(self) -> str:
         return _repr_html.fmin(self)  # type:ignore
 
-    def _repr_pretty_(self, p: Any, cycle: bool) -> None:
+    def _repr_pretty_(self, p: _tp.Any, cycle: bool) -> None:
         if cycle:
             p.text("<FMin ...>")
         else:
@@ -674,7 +678,8 @@ class Param:
     )
 
     def __init__(
-        self, *args: Union[int, str, float, Optional[Tuple[float, float]], bool]
+        self,
+        *args: _tp.Union[int, str, float, _tp.Optional[_tp.Tuple[float, float]], bool],
     ):
         """Not to be initialized by users."""
         assert len(args) == len(self.__slots__)
@@ -712,7 +717,7 @@ class Param:
         """Get user-friendly text representation."""
         return _repr_text.params([self])  # type:ignore
 
-    def _repr_pretty_(self, p: Any, cycle: bool) -> None:
+    def _repr_pretty_(self, p: _tp.Any, cycle: bool) -> None:
         if cycle:
             p.text("Param(...)")
         else:
@@ -856,7 +861,7 @@ class MError:
         "min",
     )
 
-    def __init__(self, *args: Union[int, str, float, bool]):
+    def __init__(self, *args: _tp.Union[int, str, float, bool]):
         """Not to be initialized by users."""
         assert len(args) == len(self.__slots__)
         for k, arg in zip(self.__slots__, args):
@@ -882,7 +887,7 @@ class MError:
     def _repr_html_(self) -> str:
         return _repr_html.merrors({None: self})  # type:ignore
 
-    def _repr_pretty_(self, p: Any, cycle: bool) -> None:
+    def _repr_pretty_(self, p: _tp.Any, cycle: bool) -> None:
         if cycle:
             p.text("<MError ...>")
         else:
@@ -926,8 +931,8 @@ class MErrors(OrderedDict):
 
 
 def _jacobi(
-    fn: Callable, x: np.ndarray, dx: np.ndarray, tol: float, debug: bool = False
-) -> Tuple[np.ndarray, np.ndarray]:
+    fn: _tp.Callable, x: np.ndarray, dx: np.ndarray, tol: float, debug: bool = False
+) -> _tp.Tuple[np.ndarray, np.ndarray]:
     assert x.ndim == 1
     assert dx.ndim == 1
     assert np.all(dx >= 0)
@@ -981,8 +986,10 @@ def _jacobi(
 
 @_deprecated.deprecated("use jacobi.propagate instead from jacobi library")
 def propagate(
-    fn: Callable, x: Indexable[float], cov: Indexable[Indexable[float]]
-) -> Tuple[np.ndarray, np.ndarray]:
+    fn: _tp.Callable,
+    x: _tp.Collection[float],
+    cov: _tp.Collection[_tp.Collection[float]],
+) -> _tp.Tuple[np.ndarray, np.ndarray]:
     """
     Numerically propagates the covariance into a new space.
 
@@ -1005,29 +1012,33 @@ def propagate(
         y is the result of fn(x)
         ycov is the propagated covariance matrix.
     """
-    x = np.atleast_1d(x)  # type:ignore
-    cov = np.atleast_2d(cov)  # type:ignore
+    vx = np.atleast_1d(x)  # type:ignore
+    if np.ndim(cov) != 2:  # type:ignore
+        raise ValueError("cov must be 2D array-like")
+    vcov = np.atleast_2d(cov)  # type:ignore
+    if vcov.shape[0] != vcov.shape[1]:
+        raise ValueError("cov must have shape (N, N)")
     tol = 1e-2
-    dx = (np.diag(cov) * tol) ** 0.5
+    dx = (np.diag(vcov) * tol) ** 0.5
     if not np.all(dx >= 0):
         raise ValueError("diagonal elements of covariance matrix must be non-negative")
-    y, jac = _jacobi(fn, x, dx, tol)
-    ycov = np.einsum("ij,kl,jl", jac, jac, cov)
+    y, jac = _jacobi(fn, vx, dx, tol)
+    ycov = np.einsum("ij,kl,jl", jac, jac, vcov)
     return y, np.squeeze(ycov) if np.ndim(y) == 0 else ycov
 
 
 class _Timer:
     def __init__(self, fmin):
-        self._prev = fmin.time if fmin else 0.0
+        self.value = fmin.time if fmin else 0.0
 
     def __enter__(self):
-        self.value = monotonic()
+        self.value += monotonic()
 
     def __exit__(self, *args):
-        self.value = monotonic() - self.value + self._prev
+        self.value = monotonic() - self.value
 
 
-def make_func_code(params: Indexable[str]) -> Namespace:
+def make_func_code(params: _tp.Collection[str]) -> Namespace:
     """
     Make a func_code object to fake a function signature.
 
@@ -1041,8 +1052,8 @@ def make_func_code(params: Indexable[str]) -> Namespace:
 
 
 def make_with_signature(
-    callable: Callable, *varnames: str, **replacements: str
-) -> Callable:
+    callable: _tp.Callable, *varnames: str, **replacements: str
+) -> _tp.Callable:
     """
     Return new callable with altered signature.
 
@@ -1057,18 +1068,16 @@ def make_with_signature(
     -------
     callable with new argument names.
     """
-    from typing import Tuple
-
     if replacements:
-        vars = describe(callable)
-        if vars:
+        d = describe(callable)
+        if d:
             n = len(varnames)
-            if n > len(vars):
+            if n > len(d):
                 raise ValueError("varnames longer than original signature")
-            vars[:n] = varnames
+            d[:n] = varnames
         for k, v in replacements.items():
-            vars[vars.index(k)] = v
-        vars = tuple(vars)
+            d[d.index(k)] = v
+        vars = tuple(d)
     else:
         vars = varnames
 
@@ -1078,7 +1087,7 @@ def make_with_signature(
             raise ValueError("number of parameters do not match")
 
     class Caller:
-        def __init__(self, varnames: Tuple[str, ...]):
+        def __init__(self, varnames: _tp.Tuple[str, ...]):
             self.func_code = make_func_code(varnames)  # type:ignore
 
         def __call__(self, *args: object) -> object:
@@ -1088,8 +1097,8 @@ def make_with_signature(
 
 
 def merge_signatures(
-    callables: Iterable[Callable],
-) -> Tuple[List[str], List[Tuple[int, ...]]]:
+    callables: _tp.Iterable[_tp.Callable],
+) -> _tp.Tuple[_tp.List[str], _tp.List[_tp.Tuple[int, ...]]]:
     """
     Merge signatures of callables with positional arguments.
 
@@ -1133,7 +1142,7 @@ def merge_signatures(
     return args, mapping
 
 
-def describe(callable: Callable) -> List[str]:
+def describe(callable: _tp.Callable) -> _tp.List[str]:
     """
     Attempt to extract the function argument names.
 
@@ -1199,7 +1208,7 @@ def describe(callable: Callable) -> List[str]:
     )
 
 
-def _arguments_from_func_code(obj: Any) -> List[str]:
+def _arguments_from_func_code(obj: _tp.Any) -> _tp.List[str]:
     # Check (faked) f.func_code; for backward-compatibility with iminuit-1.x
     if hasattr(obj, "func_code"):
         fc = obj.func_code
@@ -1207,7 +1216,7 @@ def _arguments_from_func_code(obj: Any) -> List[str]:
     return []
 
 
-def _arguments_from_inspect(obj: Callable) -> List[str]:
+def _arguments_from_inspect(obj: _tp.Callable) -> _tp.List[str]:
     try:
         # fails for builtin on Windows and OSX in Python 3.6
         signature = inspect.signature(obj)
@@ -1226,7 +1235,7 @@ def _arguments_from_inspect(obj: Callable) -> List[str]:
     return args
 
 
-def _arguments_from_docstring(obj: Callable) -> List[str]:
+def _arguments_from_docstring(obj: _tp.Callable) -> _tp.List[str]:
     doc = inspect.getdoc(obj)
 
     if doc is None:
@@ -1296,17 +1305,17 @@ def _arguments_from_docstring(obj: Callable) -> List[str]:
 
 
 def _guess_initial_step(val: float) -> float:
-    return 1e-2 * val if val != 0 else 1e-1  # heuristic
+    return 1e-2 * abs(val) if val != 0 else 1e-1  # heuristic
 
 
-def _key2index_from_slice(var2pos: Dict[str, int], key: slice) -> List[int]:
+def _key2index_from_slice(var2pos: _tp.Dict[str, int], key: slice) -> _tp.List[int]:
     start = var2pos[key.start] if isinstance(key.start, str) else key.start
     stop = var2pos[key.stop] if isinstance(key.stop, str) else key.stop
     start, stop, step = slice(start, stop, key.step).indices(len(var2pos))
     return list(range(start, stop, step))
 
 
-def _key2index_item(var2pos: Dict[str, int], key: Union[str, int]) -> int:
+def _key2index_item(var2pos: _tp.Dict[str, int], key: _tp.Union[str, int]) -> int:
     if isinstance(key, str):
         return var2pos[key]
     i = key
@@ -1318,18 +1327,22 @@ def _key2index_item(var2pos: Dict[str, int], key: Union[str, int]) -> int:
 
 
 def _key2index(
-    var2pos: Dict[str, int], key: Union[slice, Iterable[Union[str, int]], str, int]
-) -> Union[int, List[int]]:
-    from typing import Iterable
+    var2pos: _tp.Dict[str, int],
+    key: _mtp.Key,
+) -> _tp.Union[int, _tp.List[int]]:
+    import typing as _tp
 
     if isinstance(key, slice):
         return _key2index_from_slice(var2pos, key)
-    if not isinstance(key, str) and isinstance(key, Iterable):
+    if not isinstance(key, str) and isinstance(key, _tp.Iterable):
+        # convert boolean masks into list of indices
+        if isinstance(key[0], bool):
+            key = [k for k in range(len(var2pos)) if key[k]]
         return [_key2index_item(var2pos, k) for k in key]
     return _key2index_item(var2pos, key)
 
 
-def _address_of_cfunc(fcn: Any) -> int:
+def _address_of_cfunc(fcn: _tp.Any) -> int:
     from ctypes import c_void_p, c_double, c_uint32, POINTER, CFUNCTYPE, cast
 
     c_sig = CFUNCTYPE(c_double, c_uint32, POINTER(c_double))
@@ -1340,14 +1353,144 @@ def _address_of_cfunc(fcn: Any) -> int:
     return 0
 
 
-del Any
-del Dict
-del Union
-del Iterable
-del Callable
-del T
-del Optional
-del Sized
-del Tuple
-del Generator
-del List
+def _iterate(x):
+    if not isinstance(x, _tp.Iterable):
+        yield x
+    else:
+        for xi in x:
+            yield xi
+
+
+def _replace_none(x, v):
+    if x is None:
+        return v
+    return x
+
+
+# poor-mans progressbar if progressbar2 is not available
+class _ProgressBar:
+    def _update(self, v):
+        self._out.write(f"\r{100 * v:.0f} %")
+        self._out.flush()
+
+    def _finish(self):
+        self._out.write("\r     ")
+        self._out.flush()
+
+    def __init__(self, max_value):
+        self.max_value = max_value
+        self._out = sys.stdout
+
+        try:
+            from ipykernel.iostream import OutStream
+
+            if isinstance(self._out, OutStream):
+                from IPython.display import display, HTML
+
+                self._update = lambda v: display(
+                    HTML(
+                        f"<progress value='{100 * v:.0f}' max='100'></progress> "
+                        f"{100 * v:.0f} %"
+                    ),
+                    clear=True,
+                )
+                self._finish = lambda: display(HTML(""), clear=True)
+        except ModuleNotFoundError:
+            pass
+
+    def __enter__(self, *args):
+        self.value = 0
+        self._update(0)
+        return self
+
+    def __exit__(self, *args):
+        self._finish()
+
+    def __add__(self, v):
+        self.value += v
+        self._update(self.value / self.max_value)
+        return self
+
+
+def _histogram_segments(mask, xe, masked):
+    assert masked.ndim == 1
+
+    if mask is None:
+        return [(masked, xe)]
+
+    segments = []
+    a = 0
+    b = 0
+    am = 0
+    n = len(mask)
+    while a < n:
+        if not mask[a]:
+            a += 1
+            continue
+        b = a + 1
+        while b < n and mask[b]:
+            b += 1
+        segments.append((masked[am : am + b - a], xe[a : b + 1]))
+        am += b - a
+        a = b + 1
+    return segments
+
+
+def _show_inline_matplotlib_plots():
+    # Code taken from ipywidgets/interactive.py
+    #
+    # See comments in the original why this is needed.
+    #
+    # Copyright (c) Jupyter Development Team.
+    # Distributed under the terms of the Modified BSD License.
+    #
+    # This version was stripped down and modified to remove a deprecation warning.
+    try:
+        import matplotlib as mpl
+        from matplotlib_inline.backend_inline import flush_figures
+    except ImportError:  # pragma: no cover
+        return  # pragma: no cover
+
+    if (
+        mpl.get_backend() == "module://ipykernel.pylab.backend_inline"
+        or mpl.get_backend() == "module://matplotlib_inline.backend_inline"
+    ):
+        flush_figures()  # pragma: no cover
+
+
+def _smart_sampling(f, xmin, xmax, start=5, tol=5e-3):
+    x = np.linspace(xmin, xmax, start)
+    ynew = f(x)
+    ymin = np.min(ynew)
+    ymax = np.max(ynew)
+    y = {xi: yi for (xi, yi) in zip(x, ynew)}
+    a = x[:-1]
+    b = x[1:]
+    while len(a):
+        if len(y) > 10000:
+            warnings.warn("Too many points", RuntimeWarning)  # pragma: no cover
+            break  # pragma: no cover
+        xnew = 0.5 * (a + b)
+        ynew = f(xnew)
+        ymin = min(ymin, np.min(ynew))
+        ymax = max(ymax, np.max(ynew))
+        for xi, yi in zip(xnew, ynew):
+            y[xi] = yi
+        yint = 0.5 * (
+            np.fromiter((y[ai] for ai in a), float)
+            + np.fromiter((y[bi] for bi in b), float)
+        )
+        dy = np.abs(ynew - yint)
+
+        mask = dy > tol * (ymax - ymin)
+
+        # intervals which do not pass interpolation test
+        a = a[mask]
+        b = b[mask]
+        xnew = xnew[mask]
+        a = np.append(a, xnew)
+        b = np.append(xnew, b)
+
+    xy = list(y.items())
+    xy.sort()
+    return np.transpose(xy)
