@@ -76,21 +76,85 @@ def test_ValueView():
     assert_equal(v, (3, 2, 1))
 
 
+def test_FixedView_as_mask_for_other_views():
+    state = MnUserParameterState()
+    state.add("x", 1, 0.1)
+    state.add("y", 2, 0.1)
+    state.add("z", 3, 0.1)
+
+    fake_minuit = Namespace(
+        _var2pos={"x": 0, "y": 1, "z": 2},
+        _pos2var=("x", "y", "z"),
+        npar=3,
+        _last_state=state,
+        _copy_state_if_needed=lambda: None,
+    )
+
+    v = util.ValueView(fake_minuit)
+    f = util.FixedView(fake_minuit)
+    f[1] = True
+
+    assert_equal(f, [False, True, False])
+    assert_equal(v[f], [2])
+    assert_equal(v[~f], [1, 3])
+    v[f] = 5
+    assert_equal(v, [1, 5, 3])
+    v[~f] = [2, 4]
+    assert_equal(v, [2, 5, 4])
+
+
+def test_FixedView_comparison_with_broadcasting():
+    state = MnUserParameterState()
+    state.add("x", 1, 0.1)
+    state.add("y", 2, 0.1)
+    state.add("z", 3, 0.1)
+
+    fake_minuit = Namespace(
+        _var2pos={"x": 0, "y": 1, "z": 2},
+        _pos2var=("x", "y", "z"),
+        npar=3,
+        _last_state=state,
+        _copy_state_if_needed=lambda: None,
+    )
+
+    f = util.FixedView(fake_minuit)
+
+    assert_equal(f, [False, False, False])
+
+    # broadcasting
+    assert f == False
+    f[0] = True
+    assert_equal(f, [True, False, False])
+    assert f != False
+
+
 def test_Matrix():
     m = util.Matrix(("a", "b"))
-    m[:] = [[1, 1], [1, 4]]
-    assert_equal(m, ((1, 1), (1, 4)))
-    assert repr(m) == "[[1. 1.]\n [1. 4.]]"
+    m[:] = [[1, 2], [2, 8]]
+    assert_equal(m, ((1, 2), (2, 8)))
+    assert repr(m) == "[[1. 2.]\n [2. 8.]]"
     c = m.correlation()
-    assert_allclose(c, ((1.0, 0.5), (0.5, 1.0)))
-    assert m["a", "b"] == 1.0
-    assert m["a", 1] == 1.0
-    assert m[1, 1] == 4.0
-    assert_equal(m["b"], (1, 4))
+    assert_allclose(c, ((1.0, 0.5**0.5), (0.5**0.5, 1.0)))
+    assert m["a", "b"] == 2.0
+    assert m["a", 1] == 2.0
+    assert m[1, "a"] == 2.0
+    assert m[1, 1] == 8.0
+    assert_equal(m[0], [1, 2])
+    assert_equal(m["b"], (2, 8))
+    assert_equal(m[:], [[1, 2], [2, 8]])
+    assert_equal(m[:, 0], [1, 2])
+    assert_equal(m[:, 1], [2, 8])
+
+    # this swaps rows and cols
+    assert_equal(m[[1, 0]], [[8, 2], [2, 1]])
 
     m *= 2
-    assert_equal(m, ((2, 2), (2, 8)))
-    assert_allclose(np.dot(m, (1, 1)), (4, 10))
+    assert_equal(m, ((2, 4), (4, 16)))
+
+    m2 = np.dot(m, (1, 1))
+    assert repr(m2) == "[ 6. 20.]"
+    assert str(m2) == "[ 6. 20.]"
+    assert_allclose(m2, (6, 20))
 
     # matrix is always square
 
@@ -257,11 +321,12 @@ def test_MErrors():
     assert repr(mes) == f"<MErrors\n  {mes['x']!r}\n>"
 
 
-def test_FMin():
+@pytest.mark.parametrize("errordef", (0.5, 1.0))
+def test_FMin(errordef):
     fm = Namespace(
         fval=1.23456e-10,
         edm=1.23456e-10,
-        errordef=0.5,
+        errordef=errordef,
         is_valid=True,
         has_valid_parameters=True,
         has_accurate_covar=True,
@@ -307,13 +372,19 @@ def test_FMin():
     assert fmin != util.FMin(fm, "bar", 1, 2, 1, 0.1, 1.2)
     assert fmin != util.FMin(fm, "foo", 1, 2, 1, 0.1, 1.5)
 
+    if errordef == 1:
+        reduced_chi2 = fmin.fval
+    else:
+        reduced_chi2 = np.nan
+
     assert repr(fmin) == (
-        "<FMin algorithm='foo' edm=1.23456e-10 edm_goal=0.1 errordef=0.5 fval=1.23456e-10"
+        f"<FMin algorithm='foo' edm=1.23456e-10 edm_goal=0.1 errordef={errordef}"
+        " fval=1.23456e-10"
         " has_accurate_covar=True has_covariance=True has_made_posdef_covar=False"
         " has_parameters_at_limit=False has_posdef_covar=True"
         " has_reached_call_limit=False has_valid_parameters=True"
         " hesse_failed=False is_above_max_edm=False is_valid=True"
-        " nfcn=1 ngrad=2 reduced_chi2=2.46912e-10 time=1.2>"
+        f" nfcn=1 ngrad=2 reduced_chi2={reduced_chi2} time=1.2>"
     )
 
 
@@ -453,7 +524,7 @@ def test_propagate_1():
     np.testing.assert_allclose(ycov, [[4, 0.4, 0.8], [0.4, 8, 1.2], [0.8, 1.2, 12]])
 
     with pytest.warns(np.VisibleDeprecationWarning):
-        y, ycov = util.propagate(fn, 1, 2)
+        y, ycov = util.propagate(fn, [1], [[2]])
     np.testing.assert_allclose(y, 3)
     np.testing.assert_allclose(ycov, 8)
 
@@ -515,6 +586,15 @@ def test_propagate_on_bad_input():
         with pytest.raises(ValueError):
             util.propagate(fn, x, cov)
 
+    with pytest.warns(np.VisibleDeprecationWarning):
+        with pytest.raises(ValueError):
+            util.propagate(fn, x, 1)
+
+    cov = [[1.0], [1.0]]
+    with pytest.warns(np.VisibleDeprecationWarning):
+        with pytest.raises(ValueError):
+            util.propagate(fn, x, cov)
+
 
 def test_jacobi():
     n = 100
@@ -610,3 +690,95 @@ def test_jacobi_divergence_2(capsys):
 
     np.testing.assert_allclose(y, 1e10)
     np.testing.assert_equal(jac, np.nan)
+
+
+def test_iterate():
+    assert list(util._iterate(1)) == [1]
+    assert list(util._iterate([1, 2])) == [1, 2]
+
+
+def test_replace_none():
+    assert util._replace_none(None, 2) == 2
+    assert util._replace_none(3, 2) == 3
+
+
+def test_progressbar(capsys):
+    with util._ProgressBar(max_value=4) as bar:
+        for i in range(4):
+            bar += 1
+    stdout, stderr = capsys.readouterr()
+    assert stdout == "\r0 %\r25 %\r50 %\r75 %\r100 %\r     "
+
+
+def test_progressbar_2(capsys):
+    import sys
+
+    m_iostream = pytest.importorskip("ipykernel.iostream")
+    m_display = pytest.importorskip("IPython.display")
+
+    class OutStream:
+        def write(self, s):
+            original_stdout.write(s)
+
+        def flush(self):
+            original_stdout.flush()
+
+    def display(msg, **kwargs):
+        sys.stdout.write(msg._repr_html_())
+
+    original_stdout = sys.stdout
+    sys.stdout = OutStream()
+
+    # monkey-patching our mockups
+    m_iostream.OutStream = OutStream
+    m_display.display = display
+
+    with util._ProgressBar(max_value=4) as bar:
+        for i in range(4):
+            bar += 1
+
+    sys.stdout = original_stdout
+    stdout, stderr = capsys.readouterr()
+
+    assert stdout == (
+        "<progress value='0' max='100'></progress> 0 %"
+        "<progress value='25' max='100'></progress> 25 %"
+        "<progress value='50' max='100'></progress> 50 %"
+        "<progress value='75' max='100'></progress> 75 %"
+        "<progress value='100' max='100'></progress> 100 %"
+    )
+
+
+@pytest.mark.parametrize(
+    "mask_expected",
+    (
+        (None, [[0, 1, 2]]),
+        ([True, False], [[0]]),
+        ([False, True], [[1]]),
+        ([True, True, True, False, True, True, False, True], [[0, 1, 2], [4, 5], [7]]),
+    ),
+)
+def test_histogram_segments(mask_expected):
+    mask, expected = mask_expected
+    if mask is None:
+        xe = np.linspace(0, 1, 4)
+        masked = np.arange(3)
+    else:
+        xe = np.linspace(0, 1, len(mask))
+        masked = np.arange(len(mask))[np.array(mask)]
+    segments = util._histogram_segments(mask, xe, masked)
+    assert_equal([s[0] for s in segments], expected)
+
+
+@pytest.mark.parametrize(
+    "fn_expected", ((lambda x: x, 15), (lambda x: x**11, 40), (np.log, 80))
+)
+def test_smart_sampling_1(fn_expected):
+    fn, expected = fn_expected
+    x, y = util._smart_sampling(fn, 1e-10, 5)
+    assert len(y) < expected
+
+
+def test_smart_sampling_2():
+    with pytest.warns(RuntimeWarning):
+        util._smart_sampling(np.log, 1e-10, 1, tol=1e-10)
